import typing as tp
import numpy as np

class PIController:
    def __init__(self, Kp: float, Ki: float, integral_max: float):
        """
        A PI controller, with anti-windup
        The output of the controller is given by:
        u = - Kp * ((e) + sat(Ki * int(e)))
        where e = x - x_target is the error ; int is a time integral
        and sat a saturation at integral_max
        Parameters:
         - Kp: proportional gain
         - Ki: integral gain
         - integral_max: anti-windup: maximum value of Ki * integral
        """
        self.Kp = Kp
        self.Ki = Ki
        self.integral_max = integral_max
        self.integral = 0

    def reset_integral(self, value: float = 0):
        """
        Reset the integral to a specific value
        """
        self.integral = value

    def compute(self, e: float, dt: float):
        """
        Compute next PI output
        Parameters:
         - e: current error (x - x_target)
         - dt: time (in s) since last call
        Returns: PI output
        """
        # Anti-windup
        if self.Ki > 1e-10:
            self.integral = np.maximum(-self.integral_max / self.Ki, np.minimum(self.integral_max / self.Ki, self.integral + dt * e))

        return - self.Kp * (e + self.Ki * self.integral)



import numpy as np

def clarke_park_inv(theta: float, Vdq: np.array):
    '''
    Clarke-Park inverse transform

    @param theta Electrical angle
    @param Vdq [Vd Vq] array
    @return [Va Vb Vc] array
    '''
    C = 3 / 2 * np.array([[2 / 3, 0], [- 1 / 3, np.sqrt(3) / 3], [- 1 / 3, - np.sqrt(3) / 3]])
    P = np.array([[np.cos(theta), - np.sin(theta)],[ np.sin(theta), np.cos(theta)]])
    return C @ P @ Vdq

def clarke_park(theta, Vphase):
    '''
    Clarke-Park direct transform

    @param theta Electrical angle
    @param Vphase [Va Vb Vc] array
    @return [Vd Vq] array
    '''
    C = 2 / 3 * np.array([[1, - 1/2, - 1/2], [0, np.sqrt(3) / 2, - np.sqrt(3) / 2]])
    P = np.array([[np.cos(theta), np.sin(theta)],[ -np.sin(theta), np.cos(theta)]])
    return P @ C @ Vphase


def svpwm(theta_el: float, Vdq: np.array, Vdc: float):
    """
    Space-vector PWM to compute phase voltage from direct/quadrature vectors.

    This code is derived from the version presented in the SimpleFOC library,
    https://docs.simplefoc.com/foc_theory
    """

    # Compute amplitude and angle - clamping as needed
    Uout = np.linalg.norm(Vdq) / Vdc * np.sqrt(3)
    Uout = min(Uout, 1)

    angle = (theta_el + np.arctan2(Vdq[1], Vdq[0])) % (2 * np.pi)

    # Find the sector
    sector = np.floor(angle / (np.pi / 3.0)) + 1

    # Calculate duty cycles
    T1 = np.sqrt(3) * np.sin(sector * np.pi / 3 - angle) * Uout
    T2 = np.sqrt(3) * np.sin(angle - (sector - 1) * np.pi / 3) * Uout
    T0 = 1 - T1 - T2

    if sector == 1:
        Ta = T1 + T2 + T0/2
        Tb = T2 + T0/2
        Tc = T0/2
    elif sector == 2:
          Ta = T1 +  T0/2
          Tb = T1 + T2 + T0/2
          Tc = T0/2
    elif sector == 3:
          Ta = T0/2
          Tb = T1 + T2 + T0/2
          Tc = T2 + T0/2
    elif sector == 4:
          Ta = T0/2
          Tb = T1+ T0/2
          Tc = T1 + T2 + T0/2
    elif sector == 5:
          Ta = T2 + T0/2
          Tb = T0/2
          Tc = T1 + T2 + T0/2
    else:
          Ta = T1 + T2 + T0/2
          Tb = T0/2
          Tc = T1 + T0/2

    # Calculate the phase voltages, recentering them
    average = (Ta + Tb + Tc) / 3.0
    Ua = (Ta - average) * Vdc / np.sqrt(3)
    Ub = (Tb - average) * Vdc / np.sqrt(3)
    Uc = (Tc - average) * Vdc / np.sqrt(3)
    return np.array([Ua, Ub, Uc])


import numpy as np
import sys

class AbstractSignal:
    """
    Abstract class representing a signal
    """
    def __init__(self, frequency: float = 0, phase_shift: float = 0, amplitude: float = 0, offset: float = 0):
        self.A = amplitude
        self.offset = offset
        self.omega = 2 * np.pi * frequency
        self.phi = phase_shift

    def value(self, t: float):
        """
        Return the value of the signal at time t
        """
        return 0.0

    def derivative(self, t: float):
        """
        Return the derivative of the signal at time t
        """
        return 0.0

class SignalConstant(AbstractSignal):
    def value(self, t: float):
        return self.offset + 0 * t

class SignalSinus(AbstractSignal):
    def value(self, t: float):
        """
        Return the value of the signal at time t
        """
        return self.offset + self.A * np.sin(self.omega * t + self.phi)

    def derivative(self, t: float):
        """
        Return the derivative of the signal at time t
        """
        return self.A * self.omega * np.cos(self.omega * t + self.phi)

class SignalSquare(AbstractSignal):
    def value(self, t: float):
        """
        Return the value of the signal at time t
        """
        # Note: we are not using scipy to reduce dependencies, hence the size of
        # the Windows binary
        x = np.sign(np.sin(self.omega * t + self.phi)) / 2 + 0.5
        return self.offset + self.A * x

class SignalTriangle(AbstractSignal):
    def __init__(self, frequency: float = 0, phase_shift: float = 0, amplitude: float = 0, offset: float = 0):
        super().__init__(frequency, phase_shift, amplitude, offset)
        self.deriv = SignalSquare(frequency, phase_shift, 4 * amplitude, -2 * amplitude)
        self.f = frequency

    def value(self, t: float):
        """
        Return the value of the signal at time t
        """
        # Note: we are not using scipy to reduce dependencies, hence the size of
        # the Windows binary
        x = self.f * t + self.phi / 2 / np.pi
        return self.offset + self.A * 2 * np.abs(x - np.floor(x + 0.5))

    def derivative(self, t: float):
        """
        Return the derivative of the signal at time t
        """
        return self.deriv.value(t)

def create_signal(signal_class_name: str,
                  frequency: float,
                  phase_shift: float,
                  amplitude: float,
                  offset: float):
    '''
    Create a signal from a class name and phase values
    '''
    return getattr(sys.modules[__name__], signal_class_name)(frequency, phase_shift, amplitude, offset)



from logging.handlers import TimedRotatingFileHandler
import typing as tp
import numpy as np
from matplotlib.figure import Figure
from matplotlib.gridspec import GridSpec
import math

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk, GLib, GdkPixbuf

from .widget_motor_creation import DisplayMotor
from .abstract_tab import AbstractTab
from .widget_signal_config import SignalConfigWidget
from .widget_pi_config import PIConfigWidget
from .utils import *
from ..ressources import get_ressource_path

from ..simulation.pi_controller import PIController
from ..simulation.simulate import simulate, ControlType


def plot_position_tracking(ax, simulation_result):
    ax.plot(simulation_result.time, simulation_result.theta, label = "Mechanical angle")
    if simulation_result.control_type == ControlType.POSITION:
        ax.plot(simulation_result.time, simulation_result.pos_target, label = "Target angle")
    ax.set_ylabel("Position (rad)")
    ax.grid()
    ax.legend()

def plot_velocity_tracking(ax, simulation_result):
    ax.plot(simulation_result.time, simulation_result.dtheta, label = "Velocity")
    if simulation_result.control_type == ControlType.POSITION or simulation_result.control_type == ControlType.VELOCITY:
        ax.plot(simulation_result.time, simulation_result.vel_target, label = "Target velocity")
    ax.set_ylabel("Velocity (rad/s)")
    ax.grid()
    ax.legend()

def plot_idq(ax, simulation_result):
    ax.plot(simulation_result.time, simulation_result.idq[1], label = "Quadrature current")
    ax.plot(simulation_result.time, simulation_result.idq_target[1], label = "Quadrature current target")
    ax.plot(simulation_result.time, simulation_result.idq[0], label = "Direct current")
    ax.plot(simulation_result.time, simulation_result.idq_target[0], label = "Direct current target")
    ax.set_ylabel("Current (A)")
    ax.axhline(-simulation_result.motor.iq_max, color="k", linestyle="dashed")
    ax.axhline(simulation_result.motor.iq_max, color="k", linestyle="dashed")
    ax.set_ylim(-1.05 * simulation_result.motor.iq_max, 1.05 * simulation_result.motor.iq_max)
    ax.grid()
    ax.legend()

def plot_vdq(ax, simulation_result):
    ax.plot(simulation_result.time, simulation_result.Vdq_target[1], color="C2", label = "Quadrature voltage target")
    ax.plot(simulation_result.time, simulation_result.Vdq_target[0], color="C3", label = "Direct voltage target")
    ax.plot(simulation_result.time, simulation_result.Vdq[1], label = "Quadrature voltage")
    ax.plot(simulation_result.time, simulation_result.Vdq[0], label = "Direct voltage")
    ax.set_ylabel("Voltage (V)")
    um = simulation_result.motor.U / np.sqrt(3)
    ax.axhline(-um, color="k", linestyle="dashed")
    ax.axhline(um, color="k", linestyle="dashed")
    ax.set_ylim(-1.05 * um, 1.05 * um)
    ax.grid()
    ax.legend()

def plot_iphase(ax, simulation_result):
    ax.plot(simulation_result.time, simulation_result.iphase[0], label = "Current phase A")
    ax.plot(simulation_result.time, simulation_result.iphase[1], label = "Current phase B")
    ax.plot(simulation_result.time, simulation_result.iphase[2], label = "Current phase C")
    ax.set_ylabel("Current (A)")
    ax.axhline(-simulation_result.motor.iq_max, color="k", linestyle="dashed")
    ax.axhline(simulation_result.motor.iq_max, color="k", linestyle="dashed")
    ax.set_ylim(-1.05 * simulation_result.motor.iq_max, 1.05 * simulation_result.motor.iq_max)
    ax.grid()
    ax.legend()

def plot_uphase(ax, simulation_result):
    ax.plot(simulation_result.time, simulation_result.Vphase[0], label = "Voltage phase A")
    ax.plot(simulation_result.time, simulation_result.Vphase[1], label = "Voltage phase B")
    ax.plot(simulation_result.time, simulation_result.Vphase[2], label = "Voltage phase C")
    ax.set_ylabel("Voltage (V)")
    um = simulation_result.motor.U / np.sqrt(3)
    ax.axhline(-um, color="k", linestyle="dashed")
    ax.axhline(um, color="k", linestyle="dashed")
    ax.set_ylim(-1.05 * um, 1.05 * um)
    ax.grid()
    ax.legend()

SIMULATION_PLOTS = [("Position tracking", True, plot_position_tracking),
                    ("Velocity tracking", True, plot_velocity_tracking),
                    ("Current tracking", True, plot_idq),
                    ("Quadrature / direct voltage", True, plot_vdq),
                    ("Phase current", True, plot_iphase),
                    ("Phase voltage", True, plot_uphase),
                   ]

class SimulateMotor(AbstractTab):
    """
    GUI tab: compare two motors
    """

    def __init__(self):
        builder = Gtk.Builder()
        builder.add_from_file(get_ressource_path("simulation_tab.glade"))
        super().__init__("Simulation", builder.get_object("box_toplevel"))

        # Get the widgets we need from the builder.
        self.spin_duration = builder.get_object("spin_duration")
        self.spin_frequency = builder.get_object("spin_frequency")
        self.spin_inertia = builder.get_object("spin_inertia")
        self.spin_nu = builder.get_object("spin_nu")
        self.combo_box_type = builder.get_object("combo_box_type")

        self.input_signal_widget = SignalConfigWidget("Input signal")

        self.mpl_fig = Figure(figsize=(5, 4), dpi=100)

        self.configure_plot(self.mpl_fig, True)
        self.result = None

        box = builder.get_object("box_basic")
        box.pack_start(self.input_signal_widget.frame, False, False, 0)
        box.pack_start(self.motor_widget.top_frame, False, False, 0)
        box.reorder_child(builder.get_object("frame_plot_option"), 3)

        # Populate plot grid
        grid = builder.get_object("grid_plot_options")
        self.plot_buttons = []
        for i, (label, active, _) in enumerate(SIMULATION_PLOTS):
            b = Gtk.CheckButton(label=label)
            b.set_active(active)
            b.connect("toggled", self.plot_config_update)

            self.plot_buttons.append(b)
            grid.attach(b, i % 2, i // 2, 1, 1)


        box = builder.get_object("box_advanced")
        self.direct_current_signal_widget = SignalConfigWidget("Direct current target value")
        box.pack_start(self.direct_current_signal_widget.frame, False, False, 0)
        self.load_signal_widget = SignalConfigWidget("Additional load torque")
        box.pack_start(self.load_signal_widget.frame, False, False, 0)
        self.current_pi_widget = PIConfigWidget("Current loop PI gains", "V/A")
        box.pack_start(self.current_pi_widget.frame, False, False, 0)
        self.velocity_pi_widget = PIConfigWidget("Velocity loop PI gains", "A/(rad/s)")
        box.pack_start(self.velocity_pi_widget.frame, False, False, 0)
        self.position_pi_widget = PIConfigWidget("Position loop PI gains", "(rad/s)/rad", "rad/s")
        box.pack_start(self.position_pi_widget.frame, False, False, 0)

        # "Reasonable" initial guess for typical applications
        self.current_pi_widget.set_gains(1.0, 100.0, 10.0)
        self.velocity_pi_widget.set_gains(5.0, 5.0, 10.0)
        self.position_pi_widget.set_gains(10.0, 0.5, 10.0)

        builder.connect_signals(self)

    def run_clicked(self, button):
        tree_iter = self.combo_box_type.get_active_iter()
        if tree_iter is None:
            return
        control_type = self.combo_box_type.get_model()[tree_iter][1]

        try:
            self.result = simulate(self.motor_widget.motor,
                               ControlType[control_type],
                               self.input_signal_widget.signal,
                               self.spin_duration.get_value(),
                               self.spin_inertia.get_value(),
                               self.spin_nu.get_value(),
                               self.current_pi_widget.controller,
                               self.velocity_pi_widget.controller,
                               self.position_pi_widget.controller,
                               self.spin_frequency.get_value(),
                               0.0,
                               self.direct_current_signal_widget.signal,
                               self.load_signal_widget.signal,
                               )
        except ArithmeticError as e:
            dialog = Gtk.MessageDialog(
                message_type=Gtk.MessageType.INFO,
                buttons=Gtk.ButtonsType.OK,
                text="Warning: an error occured during simulation !",
            )
            dialog.format_secondary_text(str(e))
            dialog.run()
            dialog.destroy()

        self.user_asked_for_update()

    def plot_config_update(self, button):
        if self.result is not None:
            self.plot_need_update()

    def update_plot(self):
        if self.result is not None:
            self.mpl_fig.clear()
            n_plots = sum([1 if b.get_active() else 0 for b in self.plot_buttons])

            y = 3 if n_plots > 3 else n_plots
            x = math.ceil(n_plots / 3)
            gs = GridSpec(x, y, left = 0.05, right = 0.95, bottom = 0.05, top = 0.95, wspace = 0.2, hspace = 0.2)
            axs  = [self.mpl_fig.add_subplot(gs[i]) for i in range(n_plots)]
            callbacks = [callback for b, (_, _, callback) in zip(self.plot_buttons, SIMULATION_PLOTS) if b.get_active()]
            for a, callback in zip(axs, callbacks):
                a.sharex(axs[0])
                callback(a, self.result)
            n = min(n_plots, 3)
            for a in axs[-n:]:
                a.set_xlabel("Time (s)")
        self.mpl_fig.canvas.draw()



import typing as tp
import numpy as np

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from ..simulation.pi_controller import PIController

from ..ressources import get_ressource_path


class PIConfigWidget:
    """
    A class for configuring a PI controller
    """
    def __init__(self, name:str, kp_unit:str, ki_unit:tp.Optional[str] = None):
        builder = Gtk.Builder()
        builder.add_from_file(get_ressource_path("pi_config_widget.glade"))
        self.frame = builder.get_object("frame")

        builder.get_object("label_title").set_label(name)
        builder.get_object("label_kp").set_label(f"Kp ({kp_unit})")
        if ki_unit is None:
            ki_unit = kp_unit.split('/')[0]
        builder.get_object("label_max").set_label(f"Maximum integral correction ({ki_unit})")

        self.spin_kp = builder.get_object("spin_kp")
        self.spin_ki = builder.get_object("spin_ki")
        self.spin_max = builder.get_object("spin_max")

        builder.connect_signals(self)
        self.input_updated()

    def input_updated(self, *args, **kargs):
        self.controller = PIController(self.spin_kp.get_value(),
                                       self.spin_ki.get_value(),
                                       self.spin_max.get_value(),
                                    )

    def set_gains(self, Kp: float, Ki: float, integral_max: float):
        '''
        Set associated PI gains.
        '''
        self.spin_kp.set_value(Kp)
        self.spin_ki.set_value(Ki)
        self.spin_max.set_value(integral_max)


import typing as tp
import numpy as np
import copy

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from ..simulation.signal import create_signal

from ..ressources import get_ressource_path

class SignalConfigWidget:
    """
    A class for configuring an input signal for simulation
    """
    def __init__(self, name:str):
        builder = Gtk.Builder()
        builder.add_from_file(get_ressource_path("signal_config_widget.glade"))
        self.frame = builder.get_object("frame")

        builder.get_object("label_title").set_label(name)

        self.spin_freq = builder.get_object("spin_frequency")
        self.spin_phase = builder.get_object("spin_phase")
        self.spin_amplitude = builder.get_object("spin_amplitude")
        self.spin_offset = builder.get_object("spin_offset")
        self.combo_box_shape = builder.get_object("combo_box_shape")

        builder.connect_signals(self)
        self.input_updated()

    def input_updated(self, *args, **kargs):
        tree_iter = self.combo_box_shape.get_active_iter()
        if tree_iter is None:
            return
        signal_shape = self.combo_box_shape.get_model()[tree_iter][1]

        self.signal = create_signal(signal_shape,
                                    self.spin_freq.get_value(),
                                    self.spin_phase.get_value(),
                                    self.spin_amplitude.get_value(),
                                    self.spin_offset.get_value(),
                                    )



import typing as tp
import numpy as np
from enum import Enum

from .signal import AbstractSignal, SignalConstant
from .pi_controller import PIController
from .space_transforms import clarke_park, clarke_park_inv, svpwm
from ..physics.motor import Motor

class ControlType(Enum):
    POSITION = 1
    VELOCITY = 2
    CURRENT = 3

class SimulationResult:
    def __init__(self, time: np.array, motor: Motor, control_type: ControlType):
        self.time = time
        self.motor = motor
        self.control_type = control_type
        l = len(time)
        self.theta = np.zeros(l)
        self.dtheta = np.zeros(l)
        self.idq = np.zeros((2, l))
        self.iphase = np.zeros((3, l))
        self.Vdq = np.zeros((2, l))
        self.Vphase = np.zeros((3, l))
        self.pos_target = np.zeros(l)
        self.vel_target = np.zeros(l)
        self.idq_target = np.zeros((2, l))
        self.Vdq_target = np.zeros((2, l))
        self.load_torque = np.zeros(l)


def bemf(theta):
    '''
    Generate BEMF vector
    '''
    return np.array([np.sin(theta), np.sin(theta - 2 * np.pi / 3), np.sin(theta + 2 * np.pi / 3)])

class MotorSimulator:
    def __init__(self,
                 motor: Motor,
                 inertia: float,
                 friction: float,
                 dt: float,
                 load_torque_signal: AbstractSignal):
        '''
        A class to simulate the motion of a brushless motor using a discrete controller

        @param motor Motor to simulate
        @param inertia Mechanical inertia
        @param friction Viscuous friction
        @param dt Step size
        @param AbstractSignal Resistive torque applied to the motor
        '''
        self.motor = motor
        self.state = np.zeros(5) # Current state: theta, dtheta, iphase
        self.I = inertia
        self.nu = friction
        self.dt = dt
        self.load = load_torque_signal
        self.t = 0

    def _dynamics(self, t, x, Vphase):
        '''
        System dynamics.

        @param x System state
        @param Vphase Phase voltage
        '''
        theta = x[0]
        dtheta = x[1]
        iphase = x[2:]
        idq = clarke_park(self.motor.np * self.motor.rho * theta, iphase)
        tau = self.motor.kt_q_art * idq[1] - self.load.value(t)
        ddtheta = (- self.nu * dtheta + tau) / self.I
        dx = np.zeros(5)
        dx[0] = dtheta
        dx[1] = ddtheta
        dx[2:] = (-self.motor.R * iphase + self.motor.ke * self.motor.rho * dtheta * bemf(self.motor.np * self.motor.rho * theta) + Vphase) / self.motor.L

        return dx

    def step(self, Vdq_target: np.array):
        '''
        Integrate system state over a timestep dt, updating the system's internal state.

        @param dt Integration length
        @param Vdq_target Direct and quadrature voltage target
        '''
        Vphase = svpwm(self.motor.np * self.motor.rho * self.state[0], Vdq_target, self.motor.U)
        self.Vphase = Vphase

        self.state += self.dt * self._dynamics(self.t, self.state, Vphase)
        self.t += self.dt


def simulate(motor: Motor,
             control_type: ControlType,
             target_signal: AbstractSignal,
             duration: float,
             system_inertia: float,
             system_friction: float,
             current_controller: PIController,
             velocity_controller: PIController = PIController(0, 0, 0),
             position_controller: PIController = PIController(0, 0, 0),
             control_loop_frequency: float = 1000,
             commutation_frequency: float = 10000,
             current_direct_target: AbstractSignal = SignalConstant(),
             load_torque_signal: AbstractSignal = SignalConstant()
             ):
    """
    Simulate the motor tracking a reference trajectory using a classical
    cascade PI controller.

    The mechanical equation is a simple load + viscuous friction:
     - I ddtheta = tau - nu dtheta

    The target signal is either a position, velocity or current target.
    The control is perform in cascade as a position PI feeding a velocity PI
    feeding a current PI, which ultimately outputs a voltage target sent to the
    motor through three PWM signal (ideal mosfets)

    Parameters:
     - motor: the motor to simulate
     - TODO

    Return: simulation result
    """
    current_controller.reset_integral(0)
    velocity_controller.reset_integral(0)
    position_controller.reset_integral(0)

    dt = 1 / control_loop_frequency
    time = np.arange(0, duration + dt, dt)
    result = SimulationResult(time, motor, control_type)

    t = 0
    if control_type == ControlType.POSITION:
        result.pos_target[0] = target_signal.value(t)
        result.vel_target[0] = target_signal.derivative(t)
    elif control_type == ControlType.VELOCITY:
        result.vel_target[0] = target_signal.value(t)
    else:
        result.idq_target[0, 1] = target_signal.value(t)
    result.idq_target[0, 0] = current_direct_target.value(t)
    result.load_torque[0] = load_torque_signal.value(t)


    simulator = MotorSimulator(motor, system_inertia, system_friction, dt, load_torque_signal)

    for i in range(1, len(time)):
        t = result.time[i]
        # Position and velocity loops, if enabled.
        target_position = 0
        target_velocity = 0
        idq_target = np.array([current_direct_target.value(t), 0.0])
        if control_type == ControlType.POSITION:
            target_position = target_signal.value(t)
            target_velocity = target_signal.derivative(t)
            vel_input = position_controller.compute(result.theta[i-1] - target_position, dt)
            idq_target[1] = velocity_controller.compute(result.dtheta[i-1] - vel_input - target_velocity, dt)
        elif control_type == ControlType.VELOCITY:
            target_velocity = target_signal.value(t)
            idq_target[1] = velocity_controller.compute(result.dtheta[i-1] - target_velocity, dt)
        else:
            idq_target[1] = target_signal.value(t)

        # Saturate current target, giving priority to the quadrature current.
        idq_target[1] = min(motor.iq_max, max(-motor.iq_max, idq_target[1]))
        id_max = np.sqrt(motor.iq_max**2 - idq_target[1]**2)
        idq_target[0] = min(id_max, max(-id_max, idq_target[0]))

        Vdq_target = current_controller.compute(result.idq[:, i - 1] - idq_target, dt)

        # Integrate
        simulator.step(Vdq_target)

        # Store results
        result.theta[i] = simulator.state[0]
        result.dtheta[i] = simulator.state[1]
        result.idq[:, i] = clarke_park(motor.np * motor.rho * result.theta[i], simulator.state[2:])
        result.iphase[:, i] = simulator.state[2:]
        result.Vdq[:, i] = clarke_park(motor.np * motor.rho * result.theta[i], simulator.Vphase)
        result.Vphase[:, i] =  simulator.Vphase
        result.pos_target[i] = target_position
        result.vel_target[i] = target_velocity
        result.idq_target[:, i] = idq_target
        result.Vdq_target[:, i] = Vdq_target
        result.load_torque[i] = simulator.load.value(t)

        if np.max(np.abs(simulator.state[2:])) > 10 * motor.iq_max:
            # Simulation is unstable
            raise ArithmeticError("Excessive current detected, simulation is likely numerically unstable.\n" +\
                            "Please check controller gains or increase control frequency.")
    return result



import typing as tp
import numpy as np


class Motor:
    """
    A class to represent a PMSM motor, implementing all relevant computations
    """

    @staticmethod
    def FromDict(data):
        m = Motor(1, 1, 1, 1, 1, 1, 48, 1)
        m.update_constants(
            n=data["np"],
            R=data["R"],
            L=data["L"] / 1000.0,
            ke=data["ke"],
            iq_max=data["i_quadrature_max"],
            iq_nominal=data.get("i_quadrature_nominal", data["i_quadrature_max"]),
            U=data["U"],
            reduction_ratio=data["reduction_ratio"],
        )
        return m

    def __init__(
        self,
        n: int,
        R: float,
        L: float,
        ke: float,
        iq_max: float,
        iq_nominal: float,
        U: float,
        reduction_ratio: float,
    ):
        """
        Build a PMSM motor from the fundamental parameters.
            - n: number of poles
            - R: resistor, per phase, in Ohm
            - L: phase inductance, in H
            - ke: rotor flux in a phase
            - iq_max: maximum quadrature current (assuming no defluxing) in the motor
            - U: driver voltage
            - reduction_ratio: reduction ratio
        Note:
            - magnetic saturations not modelled
            - salliancy ratio is set to 1
        """
        self.update_constants(n, R, L, ke, iq_max, iq_nominal, U, reduction_ratio)

    def to_dict(self):
        """
        Store motore in dictionnary to save to json file.
        """
        return {
            "np": 2 * self.np,
            "R": self.R,
            "L": self.L,
            "L": 1000.0 * self.L,
            "ke": self.ke,
            "i_quadrature_max": self.iq_max,
            "i_quadrature_nominal": self.iq_nominal,
            "U": self.U,
            "reduction_ratio": self.rho,
        }

    def copy(self, other_motor: "Motor"):
        """
        Copy the constants of other_motor onto self.
        """
        self.update_constants(
            n=2 * other_motor.np,
            R=other_motor.R,
            L=other_motor.L,
            ke=other_motor.ke,
            iq_max=other_motor.iq_max,
            iq_nominal=other_motor.iq_nominal,
            U=other_motor.U,
            reduction_ratio=other_motor.rho,
        )

    def update_constants(
        self,
        n: int = None,
        R: float = None,
        L: float = None,
        ke: float = None,
        iq_max: float = None,
        iq_nominal: float = None,
        U: float = None,
        reduction_ratio: float = None,
    ):
        """
        Update the motor's constants (none to keep previous value).
        Note that updating the parameters should be done through this function
        only, otherwise the derived constants will be wrong !
        The parameters are the same as in the constructor.
        """
        if n is not None:
            self.np = n / 2
        if R is not None:
            self.R = R
        if L is not None:
            self.L = L
        if ke is not None:
            self.ke = ke
        if iq_max is not None:
            self.iq_max = iq_max
        if iq_nominal is not None:
            self.iq_nominal = iq_nominal
        if U is not None:
            self.U = U
        if reduction_ratio is not None:
            self.rho = reduction_ratio

        self._compute_derived_constants()

    def _compute_derived_constants(self):
        """
        Compute useful constants, derived from the fundamental parameters
        """
        # See BrushlessMotorPhysics.pdf for more informations on these constants.
        # Torque-related constants
        self.kt_q_art = 3.0 / 2.0 * self.rho * self.ke
        self.i_rms_max = self.iq_max / np.sqrt(2)
        self.tau_max = self.kt_q_art * self.iq_max

        # Velocity-related constants
        self.ke_phasetophase = np.sqrt(3) * self.ke
        self.w_max_no_load = self.U / self.ke_phasetophase / self.rho
        self.w_max_at_max_torque = self.compute_max_speed_no_deflux(self.tau_max)

        # Power-related constants
        self.K_m_art = np.sqrt(2.0 / 3.0) * self.kt_q_art / np.sqrt(self.R)

        # Miscellaneous.
        self.r_deflux = self.np * self.L * self.iq_max / self.ke
        tau_n = self.kt_q_art * self.iq_nominal
        self.nominal_power = self.compute_max_speed_no_deflux(tau_n) * tau_n

    def __str__(self):
        return f"R: {self.R}Ohm, L: {self.L * 1000.0}mH, Phi: {self.ke}Wb, Iq_max: {self.iq_max}A, Np {self.np}, U {self.U}, reduction {self.rho}"

    def compute_max_speed_no_deflux(self, tau):
        """
        Return the maximum articular speed, given articular torque, when not defluxing.
         - tau: input articular torque, Nm
        """
        tau = np.asarray(tau)

        iq = tau / self.kt_q_art

        a = self.rho**2 * ((self.np * self.L * iq) ** 2 + self.ke**2)
        b = 2 * self.rho * self.R * self.ke * iq
        c = (self.R * iq) ** 2 - self.U**2 / 3

        return (-b + np.sqrt(b**2 - 4 * a * c)) / 2 / a

    def compute_defluxing_current(self, tau, w):
        """
        Get defluxing current, in A, given articular torque and velocity.
        """
        tau = np.asarray(tau)
        w = np.asarray(w)

        i_q = tau / self.kt_q_art

        # Compute the minimum amount of current that is needed to deflux:
        # this corresponds to the amount of current such that U = U_eff.
        a = self.R**2 + (self.rho * w * self.np * self.L) ** 2
        b = 2 * self.np * self.L * self.ke * (self.rho * w) ** 2
        c = (
            (self.rho * w * self.np * self.L * i_q) ** 2
            + 2 * self.R * i_q * self.ke * self.rho * w
            + self.R**2 * i_q**2
            + (self.ke * self.rho * w) ** 2
            - self.U**2 / 3
        )

        return np.minimum(0.0, (-b + np.sqrt(b**2 - 4 * a * c)) / 2 / a)

    def compute_max_speed_deflux(self, tau):
        """
        Returns the maximum articular speed, given articular torque, with defluxing.
         - tau: input articular torque, Nm
        """
        tau = np.asarray(tau)

        i_q = tau / self.kt_q_art
        i_d = np.maximum(
            -np.sqrt(np.maximum(0, 2 * self.i_rms_max**2 - i_q**2)),
            -self.ke / self.np / self.L,
        )

        a = (self.rho * self.np * self.L * i_q) ** 2 + self.rho**2 * (
            self.np * self.L * i_d + self.ke
        ) ** 2
        b = 2 * self.rho * self.R * i_q * self.ke
        c = self.R**2 * (i_d**2 + i_q**2) - self.U**2 / 3
        return np.maximum(
            (-b + np.sqrt(b**2 - 4 * a * c)) / 2 / a,
            self.compute_max_speed_no_deflux(tau),
        )

    def compute_thermal_power(self, tau, w, force_no_defluxing=False):
        """
        Compute the thermal power to reach a specific working point.
        The motor will deflux if needed, unless specified.
        Note: this function does not check that the point is feasible for
        the motor: if you ask for infinite torque, you get infinite power !
        """
        tau = np.asarray(tau)
        w = np.asarray(w)

        i_q = tau / self.kt_q_art
        if force_no_defluxing:
            i_d = np.zeros(w.shape)
        else:
            i_d = self.compute_defluxing_current(tau, w)
        power = 3 / 2 * self.R * (i_d**2 + i_q**2)
        return power

    def get_power(self, w, tau):
        """
        Return the total power (mecanical + thermal) required by the motor,
        (assuming no defluxing).
        """
        return w * tau + 1 / self.K_m_art**2 * tau**2



from setuptools import setup, find_namespace_packages
from pathlib import Path

setup(
    name="nemo_bldc",
    version="1.1.0",
    description="A tool to evaluate and compare brushless motors",
    long_description=Path("README.md").read_text("utf8"),
    long_description_content_type='text/markdown',
    project_urls={
        "Source": "https://github.com/matthieuvigne/nemo_bldc",
    },
    author="Matthieu Vigne",
    license="MIT",
    packages=find_namespace_packages("src"),
    package_dir={"": "src"},
    package_data={
        "nemo_bldc.ressources": ["*.glade", "*.json", "*.svg"],
        "nemo_bldc.doc": ["*.pdf"],
    },
    install_requires=[
        "pycairo",  # Can cause issues if not installed before PyGObject - use pip and not setuptools to run install
        "pytest",
        "numpy",
        "matplotlib",
        "PyGObject",
    ],
    entry_points={"console_scripts": ["nemo_bldc = nemo_bldc.nemo:nemo_main"]},
    include_package_data=True,
    zip_safe=False,
)
